classdef PCfunction < handle
    % Piecewise Constant function on the set of semi-infinite sequences
    % determined by a complete trie and the values taken at its leaves
    %
    % author: Aurlien Garivier
    % reference: Perfect Simulation Scheme For Processes With Long Memory :
    %    A ``Coupling Into And From The Past'' Algorithm, by A. Garivier   
    
    
    properties
        CSD % Complete Suffix Trie
        values % cell of arrays 
    end
    
    methods
        function obj = PCfunction(d, v)
            % builds the function with csd = d and values = v
            % by default, returns a constant function
            % WARNING: d IS NOT COPIED! 
            if nargin<2
                d = Context.Trie(d);
                v = {0};
            end
            obj.CSD = d; % faster, be beware that d can be modified!
            %obj.CSD = d.copy(); % safer, but much slower !!
            obj.values = v;
            % obj.checkValues(); % NO USE IF THE CODE IS WELL USED
        end
        
        function f = copy(obj)
            f = Context.PCfunction(obj.CSD.copy(), obj.values);
        end        
              
        function v = getValues(obj, nodes)
            % returns the cell of values of the nodes given in the list nodes
            v = {obj.values{nodes}};        
        end
        
        function insertDetails(obj, node, subT, val)
            % modifies the function on T(node): inserts subT as a subtree
            % of node, with the values given in val
            if (subT.isLeaf(subT.getRoot()))
                obj.values{node} = val{1};
            else
                obj.CSD.replaceBySubTrie(node, subT);
                obj.values = {obj.values{:}, val{2:end}};
                obj.values{node} = [];% or: obj.checkValues(node); % or just obj.checkValues(node) % or: obj.values{node} = unique(cell2mat(val)); 
            end
        end
            
        function checkValues(obj, node) % TO REMOVE ?
            % return all values accessible from node, and update cell
            % obj.values accordingly
            if nargin<2, node=obj.CSD.getRoot(); end
            if ~obj.CSD.isLeaf(node)
                obj.values{node} = [];
                for x=1:obj.CSD.K
                    if obj.CSD.child(x, node)~=0
                        obj.checkValues(obj.CSD.child(x, node));
                    end
                end
            end
        end
        
        function newIndex = prune(obj, node)
            % prunes CSD so as to obtain the minimal csd for the given values
            % there should be a non-recursive version to avoid the Maximum
            % recursion limit of 500
            if nargin<2, node = obj.CSD.getRoot(); end
            newIndex = 1:length(obj.CSD.father);
            K = find(obj.CSD.child(:, node)~=0);
            if ~isempty(K)
                valuesAreEqual = true;
                for j=1:length(K)
                    child = obj.CSD.child(K(j), node);
                    nI = obj.prune(child); % WARNING: MODIFIES THE NODES' INDICES!
                    newIndex(newIndex>0) = nI(newIndex(newIndex>0));
                    if j==1
                        v = obj.values{child};
                    else
                        valuesAreEqual = valuesAreEqual && (obj.CSD.isLeaf(child)) && all(size(obj.values{child}) == size(v)) && all(obj.values{child} == v);
                    end
                end
                if valuesAreEqual
                    obj.values{node} = v;
                    toRemove = sort(obj.CSD.child(K, node), 1, 'descend')';
                    nI = obj.CSD.rmNodes(toRemove);
                    newIndex(newIndex>0) = nI(newIndex(newIndex>0));
                    for j=1:length(toRemove), n = toRemove(j);
                        obj.values = {obj.values{1:(n-1)}, obj.values{(n+1):end}};
                    end
                end
            end            
        end
        
        function isConst = isConstant(obj)
            % returns true if the function is constant, false otherwise
            obj.prune();
            isConst = (obj.CSD.getDepth()==0);
        end
        
        function newIndex = reorder(obj)
            % re-orders children so that, in the plot, the order of children
            % from left to right is the same as the order of labels
            % modifies obj.values  accordingly
            % returns perm: after reorder, node k is now perm(k)
            newIndex = obj.CSD.reorder();
            [obj.values{newIndex}] = obj.values{:};
        end
        
        function [x,y]=plot(obj)
            % plot of the tree; for each node n, (x(n), y(n)) in [0,1]x[0,1] is
            % the coordinate of the point corresponding to node n
            % for instance, x and y can be used like that:
            %    for j=1:length(T.father), text(x(j)+0.01, y(j),
            %    num2str(j)); end
            obj.reorder();
            [x,y] = obj.CSD.plot();
%            leaves = obj.CSD.getLeaves();
%            for j=1:length(leaves)
%                text(x(leaves(j))+0.01, y(leaves(j)), num2str(obj.values{leaves(j)})); 
%            end
            
            for j=1:length(obj.CSD.father)
                % text(x(j)+0.01, y(j), [num2str(j), ' : ', num2str(obj.values{j})]);
                text(x(j)+0.01, y(j), Context.mynum2str(obj.values{j}));
            end
        end
    end
    
    methods(Static)
        function obj = identity(K)
            % returns the identity function on {1,..., K}
            obj = Context.PCfunction(K);
            obj.CSD = Context.Trie(K); 
            obj.values = cell(1, K+1); 
            for x=1:K
                obj.CSD.addChild(1, x); 
                obj.values{x+1} = x; 
            end;
        end
    end
    
    
end

