classdef Trie < handle
    % Trees with edges labelled by integers {1,...,K}
    % Nodes are positive integers
    % Edges are stored as follows:
    %   the child of node n using edge labelled by integer x is child(x, n)
    %   the father of node n is father(n), father(trie.getRoot)=0
    %
    % author: Aurlien Garivier
    % reference: Perfect Simulation Scheme For Processes With Long Memory :
    %    A ``Coupling Into And From The Past'' Algorithm, by A. Garivier   
    
    properties
        K % alphabet size
        father % father(i) = index of the father of node i
        child % child(x, n) = child of node n following transition x
    end
    
    methods
        function obj = Trie(K, contexts)
            % constructor, alphabet of edges = {1,...,K}, default is K=2
            % builds the trie with leaves in contexts
            if nargin<1, K=2; end
            if nargin<2, contexts = {}; end
            obj.K = K;
            obj.father = 0; % initializing root
            obj.child = zeros(obj.K, 1); % initializing root
            for k=1:length(contexts), obj.addWord(contexts{k}); end
        end
        
        function T = copy(obj)
            T = Context.Trie(obj.K, {});
            T.father = obj.father;
            T.child = obj.child;
        end
        
        function father = getFather(obj, node)
            % returns the (index of) the father of node in trie
            father = obj.father(node);
        end
        
        function child = getChild(obj, node, x)
            % if node has an edge labelled by x, return the corresponding child
            % otherwise, returns 0
            child = obj.child(x, node);
        end
        
        function [T, nodes] = getSubTrie(obj, node)
            % return the trie T = subtree of obj rooted in node
            % and nodes = indices of the nodes of T in obj
            T = Context.Trie(obj.K);
            nodes = [node];
            toCheck = [node, 1];
            while ~isempty(toCheck)
                n = toCheck(1, 1); nInT = toCheck(1, 2); toCheck = toCheck(2:end, :);
                for x = 1:obj.K
                    if obj.child(x, n)>0
                        nodes = [nodes, obj.child(x, n)];
                        newChild = T.addChild(nInT, x);
                        toCheck = [toCheck; obj.child(x, n), newChild];
                    end
                end
            end
        end
        
        function word = getWord(obj, node)
            % returns the list of labels from node to the root
            n = node; word = [];
            while obj.father(n)>0 % or while n>1
                word = [word, find(obj.child(:, obj.father(n)) == n)];
                n = obj.father(n);
            end
        end
        
        function node = getNode(obj, w)
            % if there is a node in the trie labelled by w, return it with
            % otherwise, return 0
            node = obj.getRoot(); j=length(w);
            while (j>0) && (node~=0)
                node = obj.getChild(node, w(j));
            end
        end
        
        function [node, isLeaf] = getSuffix(obj, w)
            % if w has a suffix labelling a leaf of the trie, return it
            % otherwise, return 0
            node = obj.getRoot(); j=length(w);
            while (j>0) && (node~=0)
                f = node;
                node = obj.getChild(node, w(j));
                j = j-1;
            end
            if node == 0
                node = f;
                isLeaf = true;
            else
                if (j == 0)
                    isLeaf = obj.isLeaf(node);
                else
                    isLeaf = false;
                end
            end
        end
        
        function h = getDepth(obj, node)
            % returns h = depth of node in the trie
            % if node is omitted, returns the depth of the trie (ie maximal
            % distance to root)
            if nargin<2
                toCheck = [obj.getRoot(),0]; h = 0;
                while ~isempty(toCheck)
                    node = toCheck(1, 1); height = toCheck(1,2);
                    h = max(h, height);
                    toCheck = toCheck(2:end, :);
                    newToCheck = obj.child(obj.child(:, node)>0, node);
                    newToCheck(:, 2) = height+1;
                    toCheck = [toCheck; newToCheck];
                end
            else
                h = 0; n = node;
                while obj.father(n)>0 % or n>1
                    h = h + 1;
                    n = obj.father(n);
                end
            end
        end
        
        function L = getLeaves(obj, node)
            % return L = list of all (indices of) leaves in trie below node
            % (by default, node is the root of the tree)
            % L is sorted so that getLeaves(Trie(K,
            % contexts))(i) does correspond to contexts(i) 
            % non-recursive version (to avoid problems with the maximum
            % recursion limit of 500
            if nargin<2, node = obj.getRoot(); end
            toCheck = node; L = [];
            while ~isempty(toCheck)
                n = toCheck(1); toCheck = toCheck(2:end);
                if obj.isLeaf(n)
                    L = [L, n]; 
                else
                    toCheck = [toCheck; obj.child(:, n)];
                end
            end
            L = sort(L);
        end
        
        function L = getLeaves2(obj, node)
            % return L = list of all (indices of) leaves in trie below node
            % (by default, node is the root of the tree)
            % ercursive version
            if nargin<2, node = obj.getRoot(); end
            %L = find(sum(obj.child, 1)==0);
            if obj.isLeaf(node)
                L = [node]; 
            else
                L = [];
                for x=1:obj.K
                    if obj.child(x, node)~=0
                        L = [L, obj.getLeaves(obj.child(x, node))];
                    end
                end
            end
        end        
             
        function [child, new] = addChild(obj, node, x)
            % if node has an edge labelled by x, return the corresponding child
            % and new = false
            % otherwise, create it, return the new node and new = true
            if obj.child(x, node) == 0
                child = length(obj.father) + 1;
                obj.father = [obj.father, node];
                obj.child = [obj.child, zeros(obj.K, 1)];
                obj.child(x, node) = child;
                new = true;
            else % the child already exists
                child = obj.child(x, node);
                new = false;
            end
        end
        
        function [child, new] = addSubTrie(obj, node, x, T)
            % if node has no edge labelled by x, add subtree T to obj as a
            % child of (node,x) and return ok=true
            % otherwise, return the corresponding child and ok=false
            if obj.child(x, node) == 0
                offset = length(obj.father);
                obj.father = [obj.father, node, offset + T.father(2:end)];
                obj.child(x, node)=1+offset;
                obj.child = [obj.child, (T.child + offset).*(T.child>0)];
                new = true;
            else
                child = obj.child(x, node);
                new = false;
            end
        end
        
        function [node, new] = addWord(obj, w)
            % if there is a node in the trie labelled by w, return it with
            % new=false, otherwise create it and return it with new=true
            node = obj.getRoot(); new = false;
            for j=length(w):-1:1
                [node, new] = addChild(obj, node, w(j));
            end
        end
        
        function [node, ok] = replaceBySubTrie(obj, node, T)
            % if node is a leaf, replaces node by subT and returns ok=true
            % otherwise, does nothing and returns ok=false
            if obj.isLeaf(node)
                offset = length(obj.father)-1;
                nfather = T.father(2:end);
                nfather(nfather > 0) = nfather(nfather > 0) + offset;
                nfather(nfather == 1+offset) = node;
                obj.father = [obj.father, nfather];
                obj.child(:, node) = offset + T.child(:, T.getRoot());
                obj.child = [obj.child, (T.child(:, 2:end) + offset).*(T.child(:, 2:end)>0)];
                ok = true;
            else
                ok = false;
            end
        end
      
        function newIndex = rmNodes(obj, Ln)
            % removes all thes nodes (and all their childhood) of the list
            % Ln from the trie, and returns the new index numbers :
            % after reorder, node k is now newIndex(k), or if k has been
            % removed then newIndex(k) = 0;
            Ln = sort(Ln, 2, 'descend'); newIndex = 1:length(obj.father);
            %disp(['has to remove ', num2str(Ln)]);
            %if length(Ln)>0, disp([' remove ', num2str(Ln(1)), ' father is now ', num2str(obj.father)]); end;
            for k=1:length(Ln), node = Ln(k);
                nI = obj.rmNodes(obj.child(obj.child(:, node)>0, node)');
                newIndex(newIndex>0) = nI(newIndex(newIndex>0));
                newIndex(node) = 0;
                newIndex(1+node:end) = newIndex(1+node:end)-1;
                obj.father = [obj.father(1:(node-1)), obj.father((node+1):end)];
                obj.father(obj.father>node) = obj.father(obj.father>node) - 1;
                obj.child(obj.child==node) = 0;
                obj.child = obj.child-(obj.child>node);
                obj.child = [obj.child(:, 1:(node-1)), obj.child(:, (node+1):end)];
                %disp(['removed ', num2str(node), ' father is now ', num2str(obj.father)]);
            end
        end

        function isleaf = isLeaf(obj, node)
            % returns true if node is a leaf, false otherwise
            isleaf = all(obj.child(:, node)==0);
        end
        
        function complete = isComplete(obj)
            % return complete=true if the trie is complete, i.e. if every inner
            % node has exactly K children, and false otherwise
            toCheck = obj.getRoot();
            while ~isempty(toCheck)
                node = toCheck(1); toCheck = toCheck(2:end);
                if any(obj.child(:,node)>0) % if it is not a leaf...
                    if any(obj.child(:, node)==0) % if a child is missing...
                        complete = false; return; % trie incomplete
                    else % otherwise, check children
                        toCheck = [toCheck; obj.child(:, node)];
                    end
                end
            end
            complete = true;
        end
        
        function test = equals(obj, t)
            % returns true if obj is the same tree as t, and false otherwise.
            % then there is a bijection between the nodes that respects
            % childhood relationships
            n = length(obj.father);
            if length(t.father) == n
                test = true;
                corr = zeros(1,n); % corr(i) = node of obj corresponding to node i in obj
                corr(obj.getRoot) = t.getRoot;
                L = obj.getRoot;
                while ~isempty(L)
                    n = L(1); L = L(2:end);
                    alph = find(obj.child(:, n)~=0)';
                    for x=alph
                        if t.child(x, corr(n))~=0
                            corr(obj.child(x, n)) = t.child(x, corr(n));
                            L = [L, obj.child(x, n)];
                        else
                            test = false;
                            return;
                        end
                    end
                end
            else
                test = false;
            end
        end
        
        function nbAdded = complete(obj)
            % completes the trie, and returns the number of nodes added
            toCheck = obj.getRoot(); nbAdded = 0;
            while ~isempty(toCheck)
                node = toCheck(1); toCheck = toCheck(2:end);
                if any(obj.child(:,node)>0) % if it is not a leaf...
                    K = find(obj.child(:, node)==0);
                    for k=1:length(K)
                        obj.addChild(node, K(k));
                        nbAdded = nbAdded + 1;
                    end
                    K = find(obj.child(:, node)>0);
                    toCheck = [toCheck; obj.child(K, node)];
                end
            end
        end       
        
        function mute(obj)
            % mutation of the trie: a leaf's father looses its children,
            % which are given to another leaf
            leaves = obj.getLeaves();
            N = length(leaves);
            if N>2
                ok = false;
                while ~ok 
                    orig = obj.father(leaves(1+floor(rand*N)));
                    dest = leaves(1+floor(rand*N));
                    %[orig, dest]
                    n = dest; ok = true; 
                    while ok && (n ~= 0) 
                        n = obj.father(n);
                        ok = ok && (n ~= orig);
                    end
                end
                children = obj.child(:, orig);
                obj.father(children) = dest;
                obj.child(:, dest) = children;
                obj.child(:, orig) = 0;
            end
        end

        function newContexts = prefixClose(obj)
            L = obj.getLeaves();
            newContexts= [];
            for j=1:length(L), l=L(j);
                w = obj.getWord(l);
                for k=1:length(w)-1
                    [node, new] =  obj.addWord(w(1:k));
                    if new
                        newContexts = [newContexts, node];
                    end
                end
            end
            newContexts = intersect(newContexts, obj.getLeaves());
        end
        
        function newIndex = reorder(obj, node)
            % re-orders children so that, in the plot, the order of children
            % from left to right is the same as the order of labels
            % returns permutation: after reorder, node k is now newIndex(k)
            if nargin<2, node =obj.getRoot(); end
            newIndex = 1:length(obj.father);
            if any(obj.child(:,node)>0) % if it is not a leaf...
                K = find(obj.child(:, node)>0);
                for j=1:length(K), 
                    nI = obj.reorder(obj.child(K(j), node)); 
                    newIndex = nI(newIndex);
                end
                oldChild = obj.child;
                oldFather = obj.father;
                unorderedChildren = obj.child(K, node);
                orderedChildren = sort(unorderedChildren);
                for j=1:length(K) % re-affects the numbers of oldChild in the right order
                    oldNumber = unorderedChildren(j);
                    newNumber = orderedChildren(j);
                    newIndex(oldNumber) = newNumber;
                    obj.child(K(j), node) = newNumber;
                    obj.father(oldFather == oldNumber) = newNumber;
                    obj.child(:, newNumber) = oldChild(:, oldNumber);
                end
            end
        end
        
        function [x,y]=plot(obj, withNumbers)
            % plot of the tree; for each node n, (x(n), y(n)) in [0,1]x[0,1] is
            % the coordinate of the point corresponding to node n
            % if withNumbers=true, displays node numbers in the plot
            % for instance, x and y can be used like that:
            %    for j=1:length(T.father), text(x(j)+0.01, y(j), num2str(j)); end
            % To interpret the graph, it is probably a good idea to 'reorder'
            % it first, see method reorder
            if nargin<2, withNumbers = false; end
            treeplot(obj.father, 'r.', 'g');
            [x,y] = treelayout(obj.father);
            if withNumbers, for j=1:length(obj.father), text(x(j)+0.01, y(j), num2str(j)); end; end;
        end
              
        function [ok, listPb] = checkFatherChildOK(obj)
            % private function, debug purpose only
            % returns true if vectors father and child are coherent, false
            % otherwise and returns the list of problems : listPb(k, 1) has
            % an incoherent child for transition listPb(k, 2)
            ok = true; listPb = [];
            N = length(obj.father);
            if size(obj.child, 2)<N, listPb = [N, 1]; return; end
            for n = 1:N
                for x = 1:obj.K
                    if (obj.child(x, n)>0)
                        if (obj.child(x, n)>N) || (obj.father(obj.child(x, n)) ~=n)
                            ok = false;
                            listPb = [listPb; n, x]; 
                        end
                    end
                end
            end
        end
    end
    
    methods(Static)
        function root = getRoot()
            % return the index of the root
            root=1;
        end
        
        function obj = peigne(k)
            % returns a "peigne" trie (compeletely dis-balanced)
            obj = Context.Trie(2, {});
            obj.father = [0, 1+2*reshape([1 1]'*(0:k-1), [1 2*k])];
            obj.child = zeros(2, 2*k+1);
            obj.child(:, 1+2*(0:k-1)) = reshape(2:(2*k+1), [2, k]);
        end
        
        function obj = randomBinary(n)
            % returns a random trie with k leaves
            obj = Context.Trie(2);
            for k=1:(n-1)
                C = obj.getLeaves();
                j = C(1+floor(length(C)*rand));
                obj.addChild(j, 1);
                obj.addChild(j, 2);
            end
        end
        
        function L = allBinary(n)
            % return a list of all binary tries with n nodes
            if n == 1, L=Context.Trie(2, {});
            else
                L = [];
                for k = 1:(n-1) 
                    Lg = Context.Trie.allBinary(k);
                    Ld = Context.Trie.allBinary(n-k);
                    for a = 1:length(Lg)
                        for b = 1:length(Ld)
                            T = Context.Trie(2, {});
                            T.addSubTrie(T.getRoot(), 1, Lg(a));
                            T.addSubTrie(T.getRoot(), 2, Ld(b));
                            L = [L, T];
                        end
                    end
                end
            end
        end
    end
    
end